package it.asterisk.pbx.internal.activity;

import java.io.IOException;
import java.util.HashSet;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.TimeUnit;

import it.asterisk.manager.TimeoutException;
import it.asterisk.pbx.ActivityCallback;
import it.asterisk.pbx.AsteriskSettings;
import it.asterisk.pbx.Call;
import it.asterisk.pbx.CallDirection;
import it.asterisk.pbx.Channel;
import it.asterisk.pbx.EndPoint;
import it.asterisk.pbx.ListenerPriority;
import it.asterisk.pbx.PBXException;
import it.asterisk.pbx.PBXFactory;
import it.asterisk.pbx.TechType;
import it.asterisk.pbx.activities.ParkActivity;
import it.asterisk.pbx.asterisk.wrap.actions.RedirectAction;
import it.asterisk.pbx.asterisk.wrap.events.ManagerEvent;
import it.asterisk.pbx.asterisk.wrap.events.ParkedCallEvent;
import it.asterisk.pbx.asterisk.wrap.response.ManagerResponse;
import it.asterisk.pbx.internal.core.AsteriskPBX;
import it.callite.log.AsteriskLog;

/**
 * The ParkActivity is used by the AsteriksPBX implementation to park a channel.
 * The park activity expects two channels which are two legs legs of the same
 * call. The parkChannel will be redirect to the njr-park extension. The
 * hangupChannel (the second leg of the call) will be hung up. The (obvious?)
 * limitation is that we can't park something like a conference call as it has
 * more than two channels.
 * 
 * @author bsutton
 */
public class ParkActivityImpl extends ActivityHelper<ParkActivity> implements ParkActivity
{
    /**
     * call - the call which is being packed.
     */
    private final Call _call;

    /**
     * parkChannel - the channel from the above call which is going to be
     * parked. extension.
     */
    private final Channel _parkChannel;

    private volatile EndPoint _parkingLot;

    private final CountDownLatch _latch = new CountDownLatch(1);

    /*
     * @param call the that is being parked.
     * @param parkChannel the channel from the call which is the one which will
     * be parked. All other channels in the call will be hungup.
     */
    public ParkActivityImpl(final Call call, final Channel parkChannel, final ActivityCallback<ParkActivity> listener)
    {
        super("ParkActivity", listener);

        if (call == null)
        {
            throw new IllegalArgumentException("call may not be null");
        }
        if (parkChannel == null)
        {
            throw new IllegalArgumentException("parkChannel may not be null");
        }
        this._parkChannel = parkChannel;
        this._call = call;

        if (!this._call.contains(parkChannel))
            throw new IllegalArgumentException("The parkChannel must be from the call.");

        this.startActivity(true);
    }

    @Override
    public boolean doActivity() throws PBXException
    {
        boolean success = false;
        final AsteriskPBX pbx = (AsteriskPBX) PBXFactory.getActivePBX();

        AsteriskLog.debug("*******************************************************************************");
        AsteriskLog.info("***********                    begin park               ****************");
        AsteriskLog.info("***********            " + this._parkChannel + "                 ****************");
        AsteriskLog.debug("*******************************************************************************");

        try
        {
            final AsteriskSettings profile = PBXFactory.getActiveProfile();

            if (!pbx.waitForChannelToQuiescent(this._parkChannel, 3000))
                throw new PBXException("Channel: " + this._parkChannel + " cannot be parked as it is still in transition.");

            // The extension must be a non-qualified name as
            // otherwise it would be of the form DIALPLAN/njr-park
            final RedirectAction redirect = new RedirectAction(this._parkChannel, profile.getManagementContext(),
                    pbx.getExtensionPark(), 1);

            final ManagerResponse response = pbx.sendAction(redirect, 1000);
            if ((response != null) && (response.getResponse().compareToIgnoreCase("success") == 0))
            {
                // Hangup the call as we have parked the other side of
                // the call.
                if (this._call.getDirection() == CallDirection.INBOUND)
                {
                    AsteriskLog.warn("Hanging up");
                    pbx.hangup(this._call.getAcceptingParty());
                }
                else
                {
                    AsteriskLog.warn("Hanging up");
                    pbx.hangup(this._call.getOriginatingParty());
                }
                success = true;
            }

        }
        catch (IllegalArgumentException | IllegalStateException | IOException | TimeoutException e)
        {
            AsteriskLog.error(e, e);
            throw new PBXException(e);

        }
        finally
        {
            // we wait at most 2 seconds for the parking extension to
            // arrive.
            try
            {
                if (success)
                {
                    if (!this._latch.await(2, TimeUnit.SECONDS))
                    {
                        AsteriskLog.warn("Timeout, continuing");
                    }
                }

                if (this._parkingLot == null)
                {
                    success = false;
                    AsteriskLog.warn("ParkCallEvent not recieved within 2 seconds of parking call.");
                    this.setLastException(new PBXException("ParkCallEvent  not recieved within 2 seconds of parking call."));
                }
            }
            catch (final InterruptedException e)
            {
                AsteriskLog.error(e, e);

            }

        }
        return success;
    }

    @Override
    public EndPoint getParkingLot()
    {
        return this._parkingLot;
    }

    @Override
    public HashSet<Class< ? extends ManagerEvent>> requiredEvents()
    {
        HashSet<Class< ? extends ManagerEvent>> required = new HashSet<>();

        required.add(ParkedCallEvent.class);

        return required;
    }

    @Override
    synchronized public void onManagerEvent(final ManagerEvent event)
    {
        assert event instanceof ParkedCallEvent : "Unexpected event";

        final ParkedCallEvent parkedEvent = (ParkedCallEvent) event;
        final AsteriskPBX pbx = (AsteriskPBX) PBXFactory.getActivePBX();

        this._parkingLot = pbx.buildEndPoint(TechType.LOCAL, parkedEvent.getExten());
        this._latch.countDown();
    }

    @Override
    public ListenerPriority getPriority()
    {
        return ListenerPriority.NORMAL;
    }

}
