package it.callite.utility.xlsx.impl;

import static it.callite.utility.xlsx.XmlUtils.document;
import static it.callite.utility.xlsx.XmlUtils.searchForNodeList;
import static it.callite.utility.xlsx.impl.TempFileUtil.writeInputStreamToFile;
import static java.util.Arrays.asList;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.nio.file.Files;
import java.security.GeneralSecurityException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamException;

import org.apache.poi.openxml4j.exceptions.InvalidFormatException;
import org.apache.poi.openxml4j.exceptions.OpenXML4JException;
import org.apache.poi.openxml4j.opc.OPCPackage;
import org.apache.poi.poifs.crypt.Decryptor;
import org.apache.poi.poifs.crypt.EncryptionInfo;
import org.apache.poi.poifs.filesystem.POIFSFileSystem;
import org.apache.poi.ss.usermodel.Sheet;
import org.apache.poi.xssf.eventusermodel.XSSFReader;
import org.apache.poi.xssf.eventusermodel.XSSFReader.SheetIterator;
import org.apache.poi.xssf.model.SharedStrings;
import org.apache.poi.xssf.model.StylesTable;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import it.callite.utility.xlsx.StreamingReader.Builder;
import it.callite.utility.xlsx.exceptions.OpenException;
import it.callite.utility.xlsx.exceptions.ReadException;
import it.callite.utility.xlsx.sst.BufferedStringsTable;

public class StreamingWorkbookReader implements Iterable<Sheet>, AutoCloseable {

	private final List<StreamingSheet> sheets;
	private final List<Map<String, String>> sheetProperties = new ArrayList<>();
	private final Builder builder;
	private File tmp;
	private File sstCache;
	private OPCPackage pkg;
	private SharedStrings sst;
	private boolean use1904Dates = false;


	public StreamingWorkbookReader(SharedStrings sst, File sstCache, OPCPackage pkg, StreamingSheetReader reader, Builder builder) {
		this.sst = sst;
		this.sstCache = sstCache;
		this.pkg = pkg;
		this.sheets = asList(new StreamingSheet(null, reader));
		this.builder = builder;
	}

	public StreamingWorkbookReader(Builder builder) {
		this.sheets = new ArrayList<>();
		this.builder = builder;
	}

	public StreamingSheetReader first() {
		return sheets.get(0).getReader();
	}

	public void init(InputStream is) {
		File f = null;
		try {
			f = writeInputStreamToFile(is, builder.getBufferSize());
			init(f);
			tmp = f;
		} catch(IOException e) {
			throw new ReadException("Unable to read input stream", e);
		} catch(RuntimeException e) {
			if(f != null) {
				f.delete();
			}
			throw e;
		}
	}

	public void init(File f) {
		try {
			if(builder.getPassword() != null) {
				POIFSFileSystem poifs = new POIFSFileSystem(f);
				EncryptionInfo info = new EncryptionInfo(poifs);
				Decryptor d = Decryptor.getInstance(info);
				d.verifyPassword(builder.getPassword());
				pkg = OPCPackage.open(d.getDataStream(poifs));
			} else {
				pkg = OPCPackage.open(f);
			}

			XSSFReader reader = new XSSFReader(pkg);
			if(builder.getSstCacheSize() > 0) {
				sstCache = Files.createTempFile("", "").toFile();
				sst = BufferedStringsTable.getSharedStringsTable(sstCache, builder.getSstCacheSize(), pkg);
			} else {
				sst = reader.getSharedStringsTable();
			}

			StylesTable styles = reader.getStylesTable();
			NodeList workbookPr = searchForNodeList(document(reader.getWorkbookData()), "/ss:workbook/ss:workbookPr");
			if(workbookPr.getLength() == 1) {
				final Node date1904 = workbookPr.item(0).getAttributes().getNamedItem("date1904");
				if(date1904 != null) {
					use1904Dates = ("1".equals(date1904.getTextContent()));
				}
			}

			loadSheets(reader, sst, styles, builder.getRowCacheSize());
		} catch(IOException e) {
			throw new OpenException("Failed to open file", e);
		} catch(OpenXML4JException | XMLStreamException e) {
			throw new ReadException("Unable to read workbook", e);
		} catch(GeneralSecurityException e) {
			throw new ReadException("Unable to read workbook - Decryption failed", e);
		}
	}

	void loadSheets(XSSFReader reader, SharedStrings sst, StylesTable stylesTable, int rowCacheSize) throws IOException, InvalidFormatException, XMLStreamException {
		lookupSheetNames(reader);

		SheetIterator iter = (SheetIterator) reader.getSheetsData();
		Map<URI, InputStream> sheetStreams = new LinkedHashMap<>();
		while(iter.hasNext()) {
			InputStream is = iter.next();
			sheetStreams.put(iter.getSheetPart().getPartName().getURI(), is);
		}
		int i = 0;
		for(URI uri : sheetStreams.keySet()) {
			XMLEventReader parser = XMLInputFactory.newInstance().createXMLEventReader(sheetStreams.get(uri));
			sheets.add(new StreamingSheet(sheetProperties.get(i++).get("name"), new StreamingSheetReader(sst, stylesTable, parser, use1904Dates, rowCacheSize)));
		}
	}

	void lookupSheetNames(XSSFReader reader) throws IOException, InvalidFormatException {
		sheetProperties.clear();
		NodeList nl = searchForNodeList(document(reader.getWorkbookData()), "/ss:workbook/ss:sheets/ss:sheet");
		for(int i = 0; i < nl.getLength(); i++) {
			Map<String, String> props = new HashMap<>();
			props.put("name", nl.item(i).getAttributes().getNamedItem("name").getTextContent());

			Node state = nl.item(i).getAttributes().getNamedItem("state");
			props.put("state", state == null ? "visible" : state.getTextContent());
			sheetProperties.add(props);
		}
	}

	List<? extends Sheet> getSheets() {
		return sheets;
	}

	public List<Map<String, String>> getSheetProperties() {
		return sheetProperties;
	}

	@Override
	public Iterator<Sheet> iterator() {
		return new StreamingSheetIterator(sheets.iterator());
	}

	@Override
	public void close() throws IOException {
		try {
			for(StreamingSheet sheet : sheets) {
				sheet.getReader().close();
			}
			pkg.revert();
		} finally {
			if(tmp != null) {
				tmp.delete();
			}
			if(sst instanceof BufferedStringsTable) {
				((BufferedStringsTable) sst).close();
				sstCache.delete();
			}
		}
	}

	static class StreamingSheetIterator implements Iterator<Sheet> {
		private final Iterator<StreamingSheet> iterator;

		public StreamingSheetIterator(Iterator<StreamingSheet> iterator) {
			this.iterator = iterator;
		}

		@Override
		public boolean hasNext() {
			return iterator.hasNext();
		}

		@Override
		public Sheet next() {
			return iterator.next();
		}

		@Override
		public void remove() {
			throw new RuntimeException("NotSupported");
		}
	}
}
